﻿using System;
using System.Diagnostics;
using System.IO;

using Hjg.Pngcs;
using Hjg.Pngcs.Chunks;

using ProtoBuf;

namespace Automap
{
	/// <summary>
	/// Png metadata chunk.
	/// </summary>
	/// <remarks>There can be only one. (per PNG file)</remarks>
	public class PngMetadataChunk : PngChunkSingle
	{
		// ID must follow the PNG conventions: four ascii letters,
		public readonly static string ID = "cHUK";

		public ColumnMeta ChunkMetadata { get; set; }


		public PngMetadataChunk(ImageInfo info) : base(ID, info)
		{

		}

		public override ChunkOrderingConstraint GetOrderingConstraint()
		{
			return ChunkOrderingConstraint.NONE;
		}

		public override ChunkRaw CreateRawChunk()
		{
			using (MemoryStream outputStream = new MemoryStream( )) 
			{
			Serializer.Serialize<ColumnMeta>(outputStream, this.ChunkMetadata);

			ChunkRaw pngChunk = createEmptyChunk(( int )outputStream.Length, true);
			pngChunk.Data = outputStream.ToArray();

			return pngChunk;
			}
		}

		public override void ParseFromRaw(ChunkRaw pngChunk)
		{			
			using (MemoryStream inputStream = new MemoryStream(pngChunk.Data, false)) 
			{
			this.ChunkMetadata = Serializer.Deserialize<ColumnMeta>(inputStream);				
			}
		}

		public override void CloneDataFromRead(PngChunk other)
		{
			PngMetadataChunk clone = (PngMetadataChunk) other;
			this.ChunkMetadata = clone.ChunkMetadata;
		}

	}
}

